--
-- This file is part of TALER
-- Copyright (C) 2014--2023 Taler Systems SA
--
-- TALER is free software; you can redistribute it and/or modify it under the
-- terms of the GNU General Public License as published by the Free Software
-- Foundation; either version 3, or (at your option) any later version.
--
-- TALER is distributed in the hope that it will be useful, but WITHOUT ANY
-- WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
-- A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License along with
-- TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
--

CREATE TABLE IF NOT EXISTS auditor_balances
(
   row_id BIGINT GENERATED BY DEFAULT AS IDENTITY,
   balance_key TEXT PRIMARY KEY NOT NULL,
   balance_value taler_amount

);
COMMENT
ON TABLE auditor_balances
  IS 'table storing various global balances of the auditor';
COMMENT
ON COLUMN auditor_balances.balance_key
 IS 'unique name for the balance value';
COMMENT
ON COLUMN auditor_balances.balance_value
 IS 'balance amount';
