/*
   This file is part of TALER
   Copyright (C) 2022 Taler Systems SA

   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 3, or (at your option) any later version.

   TALER is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
   A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along with
   TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
 */
/**
 * @file exchangedb/pg_select_withdrawals_above_serial_id.c
 * @brief Implementation of the select_withdrawals_above_serial_id function for Postgres
 * @author Christian Grothoff
 * @author Özgür Kesim
 */
#include "platform.h"
#include "taler_error_codes.h"
#include "taler_dbevents.h"
#include "taler_pq_lib.h"
#include "pg_select_withdrawals_above_serial_id.h"
#include "pg_helper.h"

/**
 * Closure for #withdraw_serial_helper_cb().
 */
struct WithdrawSerialContext
{

  /**
   * Callback to call.
   */
  TALER_EXCHANGEDB_WithdrawCallback cb;

  /**
   * Closure for @e cb.
   */
  void *cb_cls;

  /**
   * Plugin context.
   */
  struct PostgresClosure *pg;

  /**
   * Status code, set to #GNUNET_SYSERR on hard errors.
   */
  enum GNUNET_GenericReturnValue status;
};


/**
 * Helper function to be called with the results of a SELECT statement
 * that has returned @a num_results results.
 *
 * @param cls closure of type `struct WithdrawSerialContext`
 * @param result the postgres result
 * @param num_results the number of results in @a result
 */
static void
withdraw_serial_helper_cb (void *cls,
                           PGresult *result,
                           unsigned int num_results)
{
  struct WithdrawSerialContext *rosc = cls;
  struct PostgresClosure *pg = rosc->pg;

  for (unsigned int i = 0; i<num_results; i++)
  {
    size_t num_evs;
    struct TALER_BlindedCoinHashP *h_blind_evs = NULL;
    size_t num_denom_serials;
    uint64_t *denom_serials = NULL;
    struct TALER_ReservePublicKeyP reserve_pub;
    struct TALER_ReserveSignatureP reserve_sig;
    struct GNUNET_TIME_Timestamp execution_date;
    struct TALER_HashBlindedPlanchetsP h_planchets;
    struct TALER_BlindingMasterSeedP blinding_seed;
    struct TALER_Amount amount_with_fee;
    uint64_t rowid;
    uint16_t max_age;
    uint16_t noreveal_index;
    bool no_noreveal_index;
    bool no_max_age;
    bool no_blinding_seed;
    struct GNUNET_PQ_ResultSpec rs[] = {
      TALER_PQ_result_spec_array_blinded_coin_hash (pg->conn,
                                                    "h_blind_evs",
                                                    &num_evs,
                                                    &h_blind_evs),
      GNUNET_PQ_result_spec_array_uint64 (pg->conn,
                                          "denom_serials",
                                          &num_denom_serials,
                                          &denom_serials),
      GNUNET_PQ_result_spec_auto_from_type ("reserve_pub",
                                            &reserve_pub),
      GNUNET_PQ_result_spec_auto_from_type ("reserve_sig",
                                            &reserve_sig),
      GNUNET_PQ_result_spec_auto_from_type ("h_planchets",
                                            &h_planchets),
      GNUNET_PQ_result_spec_timestamp ("execution_date",
                                       &execution_date),
      TALER_PQ_RESULT_SPEC_AMOUNT ("amount_with_fee",
                                   &amount_with_fee),
      GNUNET_PQ_result_spec_uint64 ("withdraw_id",
                                    &rowid),
      GNUNET_PQ_result_spec_allow_null (
        GNUNET_PQ_result_spec_uint16 ("max_age",
                                      &max_age),
        &no_max_age),
      GNUNET_PQ_result_spec_allow_null (
        GNUNET_PQ_result_spec_uint16 ("noreveal_index",
                                      &noreveal_index),
        &no_noreveal_index),
      GNUNET_PQ_result_spec_allow_null (
        GNUNET_PQ_result_spec_auto_from_type ("blinding_seed",
                                              &blinding_seed),
        &no_blinding_seed),
      GNUNET_PQ_result_spec_end
    };
    enum GNUNET_GenericReturnValue ret;

    if (GNUNET_OK !=
        GNUNET_PQ_extract_result (result,
                                  rs,
                                  i))
    {
      GNUNET_break (0);
      rosc->status = GNUNET_SYSERR;
      GNUNET_PQ_cleanup_result (rs);
      return;
    }
    if (num_denom_serials != num_evs)
    {
      GNUNET_break (0);
      rosc->status = GNUNET_SYSERR;
      GNUNET_PQ_cleanup_result (rs);
      return;
    }
    if (no_max_age != no_noreveal_index)
    {
      GNUNET_break (0);
      rosc->status = GNUNET_SYSERR;
      GNUNET_PQ_cleanup_result (rs);
      return;
    }

    if ((! no_max_age) &&
        ((255 <= noreveal_index) || (255 <= max_age)))
    {
      GNUNET_break (0);
      rosc->status = GNUNET_SYSERR;
      GNUNET_PQ_cleanup_result (rs);
      return;
    }
    ret = rosc->cb (rosc->cb_cls,
                    rowid,
                    num_evs,
                    h_blind_evs,
                    denom_serials,
                    &h_planchets,
                    no_blinding_seed ? NULL : &blinding_seed,
                    ! no_max_age,
                    (uint8_t) max_age,
                    (uint8_t) noreveal_index,
                    &reserve_pub,
                    &reserve_sig,
                    execution_date,
                    &amount_with_fee);
    GNUNET_PQ_cleanup_result (rs);
    if (GNUNET_OK != ret)
      break;
  }
}


enum GNUNET_DB_QueryStatus
TEH_PG_select_withdrawals_above_serial_id (
  void *cls,
  uint64_t serial_id,
  TALER_EXCHANGEDB_WithdrawCallback cb,
  void *cb_cls)
{
  struct PostgresClosure *pg = cls;
  struct GNUNET_PQ_QueryParam params[] = {
    GNUNET_PQ_query_param_uint64 (&serial_id),
    GNUNET_PQ_query_param_end
  };
  struct WithdrawSerialContext rosc = {
    .cb = cb,
    .cb_cls = cb_cls,
    .pg = pg,
    .status = GNUNET_OK
  };
  enum GNUNET_DB_QueryStatus qs;

  /* Fetch deposits with rowid '\geq' the given parameter */
  PREPARE (pg,
           "audit_get_withdraw_incr",
           "SELECT"
           " withdraw_id"
           ",h_planchets"
           ",blinding_seed"
           ",h_blind_evs"
           ",max_age"
           ",noreveal_index"
           ",denom_serials"
           ",reserve_sig"
           ",reserve_pub"
           ",execution_date"
           ",amount_with_fee"
           " FROM withdraw"
           " WHERE withdraw_id>=$1"
           " ORDER BY withdraw_id ASC;");
  qs = GNUNET_PQ_eval_prepared_multi_select (pg->conn,
                                             "audit_get_withdraw_incr",
                                             params,
                                             &withdraw_serial_helper_cb,
                                             &rosc);
  if (GNUNET_OK != rosc.status)
    return GNUNET_DB_STATUS_HARD_ERROR;
  return qs;
}
