/*
   This file is part of TALER
   Copyright (C) 2025 Taler Systems SA

   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 3, or (at your option) any later version.

   TALER is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
   A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along with
   TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
 */
/**
 * @file exchangedb/pg_disable_rules.c
 * @brief Implementation of the disable_rules function for Postgres
 * @author Christian Grothoff
 */
#include "platform.h"
#include "taler_error_codes.h"
#include "taler_dbevents.h"
#include "taler_pq_lib.h"
#include "pg_disable_rules.h"
#include "pg_helper.h"


enum GNUNET_DB_QueryStatus
TEH_PG_disable_rules (
  void *cls,
  const char *schema)
{
  struct PostgresClosure *pg = cls;
  struct GNUNET_PQ_QueryParam params[] = {
    GNUNET_PQ_query_param_string (schema),
    GNUNET_PQ_query_param_end
  };
  bool found;
  struct GNUNET_PQ_ResultSpec rs[] = {
    GNUNET_PQ_result_spec_bool ("out_found",
                                &found),
    GNUNET_PQ_result_spec_end
  };
  enum GNUNET_DB_QueryStatus qs;

  PREPARE (pg,
           "call_exchange_drop_customization",
           "SELECT out_found"
           "  FROM exchange_drop_customization"
           " ($1);");
  qs = GNUNET_PQ_eval_prepared_singleton_select (pg->conn,
                                                 "call_exchange_drop_customization",
                                                 params,
                                                 rs);
  if (qs < 0)
    return qs;
  if (found)
    return GNUNET_DB_STATUS_SUCCESS_ONE_RESULT;
  return GNUNET_DB_STATUS_SUCCESS_NO_RESULTS;
}
