/*
  This file is part of TALER
  Copyright (C) 2025 Taler Systems SA

  TALER is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation; either version 3, or (at your option) any later version.

  TALER is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  TALER; see the file COPYING.  If not, see
  <http://www.gnu.org/licenses/>
*/
/**
 * @file lib/exchange_api_melt_v27.c
 * @brief Implementation of the /melt request
 * @author Özgür Kesim
 */
#include "platform.h"
#include <jansson.h>
#include <microhttpd.h> /* just for HTTP status codes */
#include <gnunet/gnunet_util_lib.h>
#include <gnunet/gnunet_json_lib.h>
#include <gnunet/gnunet_curl_lib.h>
#include "taler_json_lib.h"
#include "taler_exchange_service.h"
#include "exchange_api_common.h"
#include "exchange_api_handle.h"
#include "taler_signatures.h"
#include "exchange_api_curl_defaults.h"
#include "exchange_api_refresh_common.h"


/**
 * @brief A /melt Handle
 */
struct TALER_EXCHANGE_MeltHandle_v27
{

  /**
   * The keys of the this request handle will use
   */
  struct TALER_EXCHANGE_Keys *keys;

  /**
   * The url for this request.
   */
  char *url;

  /**
   * The exchange base url.
   */
  char *exchange_url;

  /**
   * Curl context.
   */
  struct GNUNET_CURL_Context *cctx;

  /**
   * Context for #TEH_curl_easy_post(). Keeps the data that must
   * persist for Curl to make the upload.
   */
  struct TALER_CURL_PostContext ctx;

  /**
   * Handle for the request.
   */
  struct GNUNET_CURL_Job *job;

  /**
   * Function to call with refresh melt failure results.
   */
  TALER_EXCHANGE_MeltCallback_v27 melt_cb;

  /**
   * Closure for @e result_cb and @e melt_failure_cb.
   */
  void *melt_cb_cls;

  /**
   * Actual information about the melt operation.
   */
  struct MeltData_v27 md;

  /**
   * The secret the entire melt operation is seeded from.
   */
  struct TALER_RefreshMasterSecretP rms;

  /**
   * Details about the characteristics of the requested melt operation.
   */
  const struct TALER_EXCHANGE_MeltInput *rd;

  /**
   * True, if no blinding_seed is needed (no CS denominations involved)
   */
  bool no_blinding_seed;

  /**
   * If @e no_blinding_seed is false, the blinding seed for the intermediate
   * call to /blinding-prepare, in order to retrieve the R-values from the
   * exchange for the blind Clause-Schnorr signature.
   */
  struct TALER_BlindingMasterSeedP blinding_seed;

  /**
   * Array of `num_fresh_denom_pubs` per-coin values
   * returned from melt operation.
   */
  struct TALER_ExchangeBlindingValues *melt_blinding_values;

  /**
   * Handle for the preflight request, or NULL.
   */
  struct TALER_EXCHANGE_BlindingPrepareHandle *bpr;

  /**
   * Public key of the coin being melted.
   */
  struct TALER_CoinSpendPublicKeyP coin_pub;

  /**
   * Signature affirming the melt.
   */
  struct TALER_CoinSpendSignatureP coin_sig;

  /**
   * @brief Public information about the coin's denomination key
   */
  const struct TALER_EXCHANGE_DenomPublicKey *dki;

  /**
   * Gamma value chosen by the exchange during melt.
   */
  uint32_t noreveal_index;

};


/**
 * Verify that the signature on the "200 OK" response
 * from the exchange is valid.
 *
 * @param[in,out] mh melt handle
 * @param json json reply with the signature
 * @param[out] exchange_pub public key of the exchange used for the signature
 * @return #GNUNET_OK if the signature is valid, #GNUNET_SYSERR if not
 */
static enum GNUNET_GenericReturnValue
verify_melt_v27_signature_ok (struct TALER_EXCHANGE_MeltHandle_v27 *mh,
                              const json_t *json,
                              struct TALER_ExchangePublicKeyP *exchange_pub)
{
  struct TALER_ExchangeSignatureP exchange_sig;
  struct GNUNET_JSON_Specification spec[] = {
    GNUNET_JSON_spec_fixed_auto ("exchange_sig",
                                 &exchange_sig),
    GNUNET_JSON_spec_fixed_auto ("exchange_pub",
                                 exchange_pub),
    GNUNET_JSON_spec_uint32 ("noreveal_index",
                             &mh->noreveal_index),
    GNUNET_JSON_spec_end ()
  };

  if (GNUNET_OK !=
      GNUNET_JSON_parse (json,
                         spec,
                         NULL, NULL))
  {
    GNUNET_break_op (0);
    return GNUNET_SYSERR;
  }
  /* check that exchange signing key is permitted */
  if (GNUNET_OK !=
      TALER_EXCHANGE_test_signing_key (mh->keys,
                                       exchange_pub))
  {
    GNUNET_break_op (0);
    return GNUNET_SYSERR;
  }

  /* check that noreveal index is in permitted range */
  if (TALER_CNC_KAPPA <= mh->noreveal_index)
  {
    GNUNET_break_op (0);
    return GNUNET_SYSERR;
  }

  if (GNUNET_OK !=
      TALER_exchange_online_melt_confirmation_verify (
        &mh->md.rc,
        mh->noreveal_index,
        exchange_pub,
        &exchange_sig))
  {
    GNUNET_break_op (0);
    return GNUNET_SYSERR;
  }
  return GNUNET_OK;
}


/**
 * Function called when we're done processing the
 * HTTP /coins/$COIN_PUB/melt request.
 *
 * @param cls the `struct TALER_EXCHANGE_MeltHandle_v27`
 * @param response_code HTTP response code, 0 on error
 * @param response parsed JSON result, NULL on error
 */
static void
handle_melt_v27_finished (void *cls,
                          long response_code,
                          const void *response)
{
  struct TALER_EXCHANGE_MeltHandle_v27 *mh = cls;
  const json_t *j = response;
  struct TALER_EXCHANGE_MeltResponse_v27 mr = {
    .hr.reply = j,
    .hr.http_status = (unsigned int) response_code
  };

  mh->job = NULL;
  switch (response_code)
  {
  case 0:
    mr.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
    break;
  case MHD_HTTP_OK:
    if (GNUNET_OK !=
        verify_melt_v27_signature_ok (mh,
                                      j,
                                      &mr.details.ok.sign_key))
    {
      GNUNET_break_op (0);
      mr.hr.http_status = 0;
      mr.hr.ec = TALER_EC_EXCHANGE_MELT_INVALID_SIGNATURE_BY_EXCHANGE;
      break;
    }
    mr.details.ok.noreveal_index = mh->noreveal_index;
    mr.details.ok.num_melt_blinding_values = mh->rd->num_fresh_denom_pubs;
    mr.details.ok.melt_blinding_values = mh->melt_blinding_values;
    mr.details.ok.blinding_seed = mh->no_blinding_seed
                                               ? NULL
                                               : &mh->blinding_seed;
    mh->melt_cb (mh->melt_cb_cls,
                 &mr);
    mh->melt_cb = NULL;
    break;
  case MHD_HTTP_BAD_REQUEST:
    /* This should never happen, either us or the exchange is buggy
       (or API version conflict); just pass JSON reply to the application */
    mr.hr.ec = TALER_JSON_get_error_code (j);
    mr.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_CONFLICT:
    mr.hr.ec = TALER_JSON_get_error_code (j);
    mr.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_FORBIDDEN:
    /* Nothing really to verify, exchange says one of the signatures is
       invalid; assuming we checked them, this should never happen, we
       should pass the JSON reply to the application */
    mr.hr.ec = TALER_JSON_get_error_code (j);
    mr.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_NOT_FOUND:
    /* Nothing really to verify, this should never
       happen, we should pass the JSON reply to the application */
    mr.hr.ec = TALER_JSON_get_error_code (j);
    mr.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_INTERNAL_SERVER_ERROR:
    /* Server had an internal issue; we should retry, but this API
       leaves this to the application */
    mr.hr.ec = TALER_JSON_get_error_code (j);
    mr.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  default:
    /* unexpected response code */
    mr.hr.ec = TALER_JSON_get_error_code (j);
    mr.hr.hint = TALER_JSON_get_error_hint (j);
    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                "Unexpected response code %u/%d for exchange melt\n",
                (unsigned int) response_code,
                mr.hr.ec);
    GNUNET_break_op (0);
    break;
  }
  if (NULL != mh->melt_cb)
    mh->melt_cb (mh->melt_cb_cls,
                 &mr);
  TALER_EXCHANGE_melt_v27_cancel (mh);
}


/**
 * Start the actual melt operation, now that we have
 * the exchange's input values.
 *
 * @param[in,out] mh melt operation to run
 * @return #GNUNET_OK if we could start the operation
 */
static enum GNUNET_GenericReturnValue
start_melt (struct TALER_EXCHANGE_MeltHandle_v27 *mh)
{
  json_t *j_request_body;
  json_t *j_coin_evs;
  CURL *eh;
  struct TALER_DenominationHashP h_denom_pub;

  if (GNUNET_OK !=
      TALER_EXCHANGE_get_melt_data_v27 (&mh->rms,
                                        mh->rd,
                                        mh->no_blinding_seed
                                        ? NULL
                                        : &mh->blinding_seed,
                                        mh->melt_blinding_values,
                                        &mh->md))
  {
    GNUNET_break (0);
    return GNUNET_SYSERR;
  }
  TALER_denom_pub_hash (&mh->md.melted_coin.pub_key,
                        &h_denom_pub);
  TALER_wallet_melt_sign (
    &mh->md.melted_coin.melt_amount_with_fee,
    &mh->md.melted_coin.fee_melt,
    &mh->md.rc,
    &h_denom_pub,
    mh->md.melted_coin.h_age_commitment,
    &mh->md.melted_coin.coin_priv,
    &mh->coin_sig);
  GNUNET_CRYPTO_eddsa_key_get_public (
    &mh->md.melted_coin.coin_priv.eddsa_priv,
    &mh->coin_pub.eddsa_pub);
  mh->dki = TALER_EXCHANGE_get_denomination_key (mh->keys,
                                                 &mh->md.melted_coin.pub_key);
  j_request_body = GNUNET_JSON_PACK (
    GNUNET_JSON_pack_data_auto ("old_coin_pub",
                                &mh->coin_pub),
    GNUNET_JSON_pack_data_auto ("old_denom_pub_h",
                                &h_denom_pub),
    TALER_JSON_pack_denom_sig ("old_denom_sig",
                               &mh->md.melted_coin.sig),
    GNUNET_JSON_pack_data_auto ("confirm_sig",
                                &mh->coin_sig),
    TALER_JSON_pack_amount ("value_with_fee",
                            &mh->md.melted_coin.melt_amount_with_fee),
    GNUNET_JSON_pack_allow_null (
      (NULL != mh->md.melted_coin.h_age_commitment)
      ? GNUNET_JSON_pack_data_auto ("old_age_commitment_h",
                                    mh->md.melted_coin.h_age_commitment)
      : GNUNET_JSON_pack_string ("old_age_commitment_h",
                                 NULL)),
    GNUNET_JSON_pack_data_auto ("refresh_seed",
                                &mh->md.refresh_seed),
    GNUNET_JSON_pack_allow_null (
      (mh->md.no_blinding_seed)
      ? GNUNET_JSON_pack_string ("blinding_seed",
                                 NULL)
      : GNUNET_JSON_pack_data_auto ("blinding_seed",
                                    &mh->md.blinding_seed)),
    TALER_JSON_pack_array_of_data ("denoms_h",
                                   mh->md.num_fresh_coins,
                                   mh->md.denoms_h,
                                   sizeof(*mh->md.denoms_h))
    );
  GNUNET_assert (NULL != j_request_body);
  j_coin_evs = json_array ();
  GNUNET_assert (NULL != j_coin_evs);
  /**
   * Fill the kappa array of coin envelopes
   */
  for (uint8_t k=0; k<TALER_CNC_KAPPA; k++)
  {
    json_t *j_envs = json_array ();
    GNUNET_assert (NULL != j_envs);
    for (size_t i = 0; i < mh->md.num_fresh_coins; i++)
    {
      json_t *j_coin = GNUNET_JSON_PACK (
        TALER_JSON_pack_blinded_planchet (NULL,
                                          &mh->md.kappa_blinded_planchets[k][i]));
      GNUNET_assert (NULL != j_coin);
      GNUNET_assert (0 ==
                     json_array_append_new (j_envs, j_coin));
    }
    GNUNET_assert (0 ==
                   json_array_append_new (j_coin_evs, j_envs));
  }
  GNUNET_assert (0 ==
                 json_object_set_new (j_request_body,
                                      "coin_evs",
                                      j_coin_evs));
  /* and now we can at last begin the actual request handling */
  mh->url = TALER_url_join (mh->exchange_url,
                            "melt",
                            NULL);
  if (NULL == mh->url)
  {
    json_decref (j_request_body);
    return GNUNET_SYSERR;
  }
  eh = TALER_EXCHANGE_curl_easy_get_ (mh->url);
  if ( (NULL == eh) ||
       (GNUNET_OK !=
        TALER_curl_easy_post (&mh->ctx,
                              eh,
                              j_request_body)) )
  {
    GNUNET_break (0);
    if (NULL != eh)
      curl_easy_cleanup (eh);
    json_decref (j_request_body);
    return GNUNET_SYSERR;
  }
  json_decref (j_request_body);
  mh->job = GNUNET_CURL_job_add2 (mh->cctx,
                                  eh,
                                  mh->ctx.headers,
                                  &handle_melt_v27_finished,
                                  mh);
  return GNUNET_OK;
}


/**
 * The melt request @a mh failed, return an error to
 * the application and cancel the operation.
 *
 * @param[in] mh melt request that failed
 * @param ec error code to fail with
 */
static void
fail_mh (struct TALER_EXCHANGE_MeltHandle_v27 *mh,
         enum TALER_ErrorCode ec)
{
  struct TALER_EXCHANGE_MeltResponse_v27 mr = {
    .hr.ec = ec
  };

  mh->melt_cb (mh->melt_cb_cls,
               &mr);
  TALER_EXCHANGE_melt_v27_cancel (mh);
}


/**
 * Callbacks of this type are used to serve the result of submitting a
 * /blinding-prepare request to a exchange.
 *
 * @param cls closure with our `struct TALER_EXCHANGE_MeltHandle_v27 *`
 * @param bpr response details
 */
static void
blinding_prepare_cb (void *cls,
                     const struct TALER_EXCHANGE_BlindingPrepareResponse *bpr)
{
  struct TALER_EXCHANGE_MeltHandle_v27 *mh = cls;
  unsigned int nks_off = 0;

  mh->bpr = NULL;
  if (MHD_HTTP_OK != bpr->hr.http_status)
  {
    struct TALER_EXCHANGE_MeltResponse_v27 mr = {
      .hr = bpr->hr
    };

    mr.hr.hint = "/blinding-prepare failed";
    mh->melt_cb (mh->melt_cb_cls,
                 &mr);
    TALER_EXCHANGE_melt_v27_cancel (mh);
    return;
  }
  for (unsigned int i = 0; i<mh->rd->num_fresh_denom_pubs; i++)
  {
    const struct TALER_EXCHANGE_DenomPublicKey *fresh_pk =
      &mh->rd->fresh_denom_pubs[i];
    struct TALER_ExchangeBlindingValues *wv = &mh->melt_blinding_values[i];

    switch (fresh_pk->key.bsign_pub_key->cipher)
    {
    case GNUNET_CRYPTO_BSA_INVALID:
      GNUNET_break (0);
      fail_mh (mh,
               TALER_EC_GENERIC_CLIENT_INTERNAL_ERROR);
      return;
    case GNUNET_CRYPTO_BSA_RSA:
      break;
    case GNUNET_CRYPTO_BSA_CS:
      TALER_denom_ewv_copy (wv,
                            &bpr->details.ok.blinding_values[nks_off]);
      nks_off++;
      break;
    }
  }
  if (GNUNET_OK !=
      start_melt (mh))
  {
    GNUNET_break (0);
    fail_mh (mh,
             TALER_EC_GENERIC_CLIENT_INTERNAL_ERROR);
    return;
  }
}


struct TALER_EXCHANGE_MeltHandle_v27 *
TALER_EXCHANGE_melt_v27 (
  struct GNUNET_CURL_Context *ctx,
  const char *url,
  struct TALER_EXCHANGE_Keys *keys,
  const struct TALER_RefreshMasterSecretP *rms,
  const struct TALER_EXCHANGE_MeltInput *rd,
  TALER_EXCHANGE_MeltCallback_v27 melt_cb,
  void *melt_cb_cls)
{
  struct TALER_EXCHANGE_NonceKey nks[GNUNET_NZL (rd->num_fresh_denom_pubs)];
  unsigned int nks_off = 0;
  struct TALER_EXCHANGE_MeltHandle_v27 *mh;

  if (0 == rd->num_fresh_denom_pubs)
  {
    GNUNET_break (0);
    return NULL;
  }
  mh = GNUNET_new (struct TALER_EXCHANGE_MeltHandle_v27);
  mh->noreveal_index = TALER_CNC_KAPPA; /* invalid value */
  mh->cctx = ctx;
  mh->exchange_url = GNUNET_strdup (url);
  mh->rd = rd;
  mh->rms = *rms;
  mh->melt_cb = melt_cb;
  mh->melt_cb_cls = melt_cb_cls;
  mh->no_blinding_seed = true;
  mh->melt_blinding_values =
    GNUNET_new_array (rd->num_fresh_denom_pubs,
                      struct TALER_ExchangeBlindingValues);
  for (unsigned int i = 0; i<rd->num_fresh_denom_pubs; i++)
  {
    const struct TALER_EXCHANGE_DenomPublicKey *fresh_pk =
      &rd->fresh_denom_pubs[i];

    switch (fresh_pk->key.bsign_pub_key->cipher)
    {
    case GNUNET_CRYPTO_BSA_INVALID:
      GNUNET_break (0);
      GNUNET_free (mh->melt_blinding_values);
      GNUNET_free (mh);
      return NULL;
    case GNUNET_CRYPTO_BSA_RSA:
      TALER_denom_ewv_copy (&mh->melt_blinding_values[i],
                            TALER_denom_ewv_rsa_singleton ());
      break;
    case GNUNET_CRYPTO_BSA_CS:
      nks[nks_off].pk = fresh_pk;
      nks[nks_off].cnc_num = i;
      nks_off++;
      break;
    }
  }
  mh->keys = TALER_EXCHANGE_keys_incref (keys);
  if (0 != nks_off)
  {
    mh->no_blinding_seed = false;
    TALER_cs_refresh_secret_to_blinding_seed (
      rms,
      &mh->blinding_seed);
    mh->bpr = TALER_EXCHANGE_blinding_prepare_for_melt (ctx,
                                                        url,
                                                        &mh->blinding_seed,
                                                        nks_off,
                                                        nks,
                                                        &blinding_prepare_cb,
                                                        mh);
    if (NULL == mh->bpr)
    {
      GNUNET_break (0);
      TALER_EXCHANGE_melt_v27_cancel (mh);
      return NULL;
    }
    return mh;
  }
  if (GNUNET_OK !=
      start_melt (mh))
  {
    GNUNET_break (0);
    TALER_EXCHANGE_melt_v27_cancel (mh);
    return NULL;
  }
  return mh;
}


void
TALER_EXCHANGE_melt_v27_cancel (struct TALER_EXCHANGE_MeltHandle_v27 *mh)
{
  for (unsigned int i = 0; i<mh->rd->num_fresh_denom_pubs; i++)
    TALER_denom_ewv_free (&mh->melt_blinding_values[i]);
  if (NULL != mh->job)
  {
    GNUNET_CURL_job_cancel (mh->job);
    mh->job = NULL;
  }
  if (NULL != mh->bpr)
  {
    TALER_EXCHANGE_blinding_prepare_cancel (mh->bpr);
    mh->bpr = NULL;
  }
  TALER_EXCHANGE_free_melt_data_v27 (&mh->md); /* does not free 'md' itself */
  GNUNET_free (mh->melt_blinding_values);
  GNUNET_free (mh->url);
  GNUNET_free (mh->exchange_url);
  TALER_curl_easy_post_finished (&mh->ctx);
  TALER_EXCHANGE_keys_decref (mh->keys);
  GNUNET_free (mh);
}


/* end of exchange_api_melt.c */
